/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.portals;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.sys.funclib.debug.*;

import java.util.*;
import java.util.stream.*;

/**
 * Abstrakte Fabrik für Systemobjekte.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings({"ClassDataAbstractionCoupling", "ClassFanOutComplexity"})
public class ObjektFactory {

    private static final String TYP_STRASSENSEGMENT = "typ.straßenSegment";
    private static final String TYP_AEUSSERES_STRASSENSEGMENT = "typ.äußeresStraßenSegment";
    private static final String TYP_INNERES_STRASSENSEGMENT = "typ.inneresStraßenSegment";
    private static final String TYP_STRASSENTEILSEGMENT = "typ.straßenTeilSegment";
    private static final String TYP_MESSQUERSCHNITT = "typ.messQuerschnittAllgemein";

    private static final String TYP_UMFELDDATENMESSSTELLE = "typ.umfeldDatenMessStelle";
    private static final String TYP_UMFELDDATENSENSOR = "typ.umfeldDatenSensor";
    private static final String TYP_HELLIGKEITSSENSOR = "typ.ufdsHelligkeit";
    private static final String TYP_NIEDERSCHLAGSINTENSITAETSSENSOR = "typ.ufdsNiederschlagsIntensität";
    private static final String TYP_SICHTWEITESENSOR = "typ.ufdsSichtWeite";
    private static final String TYP_WINDGESCHWINDIGKEITSSENSOR = "typ.ufdsWindGeschwindigkeitMittelWert";
    private static final String TYP_WINDRICHTUNGSSENSOR = "typ.ufdsWindRichtung";

    private final Debug log = Debug.getLogger();
    private final Map<String, Systemobjekt> cache = new LinkedHashMap<>();
    private final DataModel model;

    public ObjektFactory(DataModel model) {
        this.model = model;
    }

    public Collection<Systemobjekt> bestimmeObjekte(Collection<String> objektPids) {
        log.info("Bestimme die zu verwaltenden Systemobjekte ...");
        Collection<SystemObject> objekte;
        if (objektPids.isEmpty()) {
            objekte = new ArrayList<>();
            objekte.addAll(model.getType(TYP_STRASSENTEILSEGMENT).getObjects());
            objekte.addAll(model.getType(TYP_MESSQUERSCHNITT).getObjects());
            objekte.addAll(model.getType(TYP_UMFELDDATENMESSSTELLE).getObjects());
            objekte.addAll(model.getType(TYP_UMFELDDATENSENSOR).getObjects());
        } else {
            objekte = model.getObjects(objektPids.toArray(new String[0])).stream()
                    .flatMap(ObjektFactory::loeseKonfigurationsbereicheUndObjekttypenAuf)
                    .filter(Objects::nonNull)
                    .distinct()
                    .collect(Collectors.toList());
        }
        return objekte.stream()
                .map(this::gibSystemobjekt)
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
    }

    private static Stream<? extends SystemObject> loeseKonfigurationsbereicheUndObjekttypenAuf(SystemObject objekt) {
        if (objekt instanceof ConfigurationArea) {
            ConfigurationArea kb = (ConfigurationArea) objekt;
            return kb.getCurrentObjects().stream();
        } else if (objekt instanceof SystemObjectType) {
            SystemObjectType typ = (SystemObjectType) objekt;
            return typ.getObjects().stream();
        } else {
            return Stream.of(objekt);
        }
    }

    Systemobjekt gibSystemobjekt(SystemObject objekt) {
        if (!cache.containsKey(objekt.getPid())) {
            cache.put(objekt.getPid(), erzeugeSystemobjekt(objekt));
        }

        return cache.get(objekt.getPid());
    }

    @SuppressWarnings("CyclomaticComplexity")
    private Systemobjekt erzeugeSystemobjekt(SystemObject objekt) {
        if (objekt.isOfType(TYP_STRASSENSEGMENT)) {
            return erzeugeStrassensegment(objekt);
        } else if (objekt.isOfType(TYP_STRASSENTEILSEGMENT)) {
            return erzeugeStrassenteilsegment(objekt);
        } else if (objekt.isOfType(TYP_MESSQUERSCHNITT)) {
            return erzeugeMessquerschnitt(objekt);
        } else if (objekt.isOfType(TYP_UMFELDDATENMESSSTELLE)) {
            return erzeugeUmfelddatenmessstelle(objekt);
        } else if (objekt.isOfType(TYP_HELLIGKEITSSENSOR)) {
            return erzeugeHelligkeitssensor(objekt);
        } else if (objekt.isOfType(TYP_NIEDERSCHLAGSINTENSITAETSSENSOR)) {
            return erzeugeNiederschlagsintensitaetssensor(objekt);
        } else if (objekt.isOfType(TYP_SICHTWEITESENSOR)) {
            return erzeugeSichtweitesensor(objekt);
        } else if (objekt.isOfType(TYP_WINDGESCHWINDIGKEITSSENSOR)) {
            return erzeugeWindgeschwindigkeitssensor(objekt);
        } else if (objekt.isOfType(TYP_WINDRICHTUNGSSENSOR)) {
            return erzeugeWindrichtungssensor(objekt);
        } else if (objekt.isOfType(TYP_UMFELDDATENSENSOR)) {
            return erzeugeUmfelddatensensor(objekt);
        } else {
            // Objekt wird nicht benötigt
            return null;
        }
    }

    private Systemobjekt erzeugeStrassensegment(SystemObject objekt) {
        Strassensegment s = new Strassensegment(objekt.getPid());
        s.setTeilsegmente(bestimmeTeilsegmente(objekt));
        s.getTeilsegmente().forEach(sts -> sts.setGehoertZu(s));
        s.setTyp(bestimmeSegmenttyp(objekt));
        s.setLaenge(bestimmeSegmentlaenge(objekt));
        return s;
    }

    private List<Strassenteilsegment> bestimmeTeilsegmente(SystemObject strassensegment) {
        List<Strassenteilsegment> teilsegmente = new ArrayList<>();
        ConfigurationObject cObject = (ConfigurationObject) strassensegment;

        AttributeGroup atg = model.getAttributeGroup("atg.bestehtAusLinienObjekten");
        Data data = cObject.getConfigurationData(atg);
        if (data != null) {
            Data.ReferenceArray linien = data.getReferenceArray("LinienReferenz");
            for (SystemObject sts : linien.getSystemObjectArray()) {
                teilsegmente.add((Strassenteilsegment) gibSystemobjekt(sts));
            }
        }
        return teilsegmente;
    }

    private Segmenttyp bestimmeSegmenttyp(SystemObject objekt) {
        if (objekt.isOfType(TYP_INNERES_STRASSENSEGMENT)) {
            return Segmenttyp.VERFLECHTUNGSBEREICH;
        } else {
            return Segmenttyp.FREIE_STRECKE;
        }
    }

    private Laenge bestimmeSegmentlaenge(SystemObject strassensegment) {
        ConfigurationObject cObjekt = (ConfigurationObject) strassensegment;

        AttributeGroup atg = model.getAttributeGroup("atg.straßenSegment");
        Data data = cObjekt.getConfigurationData(atg);
        if (data != null) {
            double laenge = data.getUnscaledValue("Länge").doubleValue();
            return Laenge.von(laenge);
        }

        atg = model.getAttributeGroup("atg.linie");
        data = cObjekt.getConfigurationData(atg);
        if (data != null) {
            double laenge = data.getUnscaledValue("Länge").doubleValue();
            return Laenge.von(laenge);
        }

        return Laenge.UNDEFINIERT;
    }

    private Systemobjekt erzeugeStrassenteilsegment(SystemObject objekt) {
        Strassenteilsegment sts = new Strassenteilsegment(objekt.getPid());
        sts.setAnzahlFahrstreifen(bestimmeAnzahlFahrstreifenVonTeilsegment(objekt));
        sts.setLaenge(bestimmeTeilsegmentlaenge(objekt));
        return sts;
    }

    private int bestimmeAnzahlFahrstreifenVonTeilsegment(SystemObject strassenteilsegment) {
        ConfigurationObject cObjekt = (ConfigurationObject) strassenteilsegment;

        AttributeGroup atg = model.getAttributeGroup("atg.straßenTeilSegment");
        Data data = cObjekt.getConfigurationData(atg);
        if (data != null) {
            return data.getScaledValue("AnzahlFahrStreifen").intValue();
        }

        return 0;
    }

    private Laenge bestimmeTeilsegmentlaenge(SystemObject strassenteilsegment) {
        ConfigurationObject cObjekt = (ConfigurationObject) strassenteilsegment;

        AttributeGroup atg = model.getAttributeGroup("atg.straßenTeilSegment");
        Data data = cObjekt.getConfigurationData(atg);
        if (data != null) {
            double laenge = data.getUnscaledValue("Länge").doubleValue();
            return Laenge.von(laenge);
        }

        atg = model.getAttributeGroup("atg.linie");
        data = cObjekt.getConfigurationData(atg);
        if (data != null) {
            double laenge = data.getUnscaledValue("Länge").doubleValue();
            return Laenge.von(laenge);
        }

        return Laenge.UNDEFINIERT;
    }

    private Messquerschnitt erzeugeMessquerschnitt(SystemObject objekt) {
        Messquerschnitt mq = new Messquerschnitt(objekt.getPid());
        mq.setLiegtAuf(bestimmeLiegtAufLinie(objekt));
        if (mq.getLiegtAuf().isPresent()) {
            Linie linie = mq.getLiegtAuf().get().getLinie();
            if (linie instanceof Strassenteilsegment) {
                Strassenteilsegment sts = (Strassenteilsegment) linie;
                sts.getMessquerschnitte().add(mq);
            } else if (linie instanceof Strassensegment) {
                mq.getStrassenteilsegment().ifPresent(sts -> sts.getMessquerschnitte().add(mq));
            }
        }
        mq.setAnzahlFahrstreifen(bestimmeAnzahlFahrstreifenVonMessquerschnitt(objekt));
        return mq;
    }

    private Linienreferenz bestimmeLiegtAufLinie(SystemObject messquerschnitt) {
        ConfigurationObject cObjekt = (ConfigurationObject) messquerschnitt;
        AttributeGroup atg = model.getAttributeGroup("atg.punktLiegtAufLinienObjekt");
        Data data = cObjekt.getConfigurationData(atg);
        if (data == null) {
            return null;
        }

        SystemObject o = data.getReferenceValue("LinienReferenz").getSystemObject();
        Linie linie = (Linie) gibSystemobjekt(o);
        double offset = data.getScaledValue("Offset").doubleValue();
        return new Linienreferenz(linie, Laenge.von(offset));
    }

    private int bestimmeAnzahlFahrstreifenVonMessquerschnitt(SystemObject messquerschnitt) {
        if (!messquerschnitt.isOfType("typ.messQuerschnitt")) {
            return 0;
        }

        ConfigurationObject cObjekt = (ConfigurationObject) messquerschnitt;
        return cObjekt.getNonMutableSet("FahrStreifen").getElements().size();
    }

    private Systemobjekt erzeugeUmfelddatenmessstelle(SystemObject objekt) {
        ConfigurationObject cObjekt = (ConfigurationObject) objekt;
        List<SystemObject> elements = cObjekt.getNonMutableSet("UmfeldDatenSensoren").getElements();
        Collection<Umfelddatensensor> sensoren = elements.stream()
                .map(ufds -> (Umfelddatensensor) gibSystemobjekt(ufds))
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
        Umfelddatenmessstelle ufdms = new Umfelddatenmessstelle(objekt.getPid());
        ufdms.setUmfelddatensensoren(sensoren);
        sensoren.forEach(ufds -> ufds.setGehoertZu(ufdms));
        return ufdms;
    }

    private Helligkeitssensor erzeugeHelligkeitssensor(SystemObject objekt) {
        return new Helligkeitssensor(objekt.getPid());
    }

    private Niederschlagsintensitaetssensor erzeugeNiederschlagsintensitaetssensor(SystemObject objekt) {
        return new Niederschlagsintensitaetssensor(objekt.getPid());
    }

    private Sichtweitesensor erzeugeSichtweitesensor(SystemObject objekt) {
        return new Sichtweitesensor(objekt.getPid());
    }

    private Windgeschwindigkeitssensor erzeugeWindgeschwindigkeitssensor(SystemObject objekt) {
        return new Windgeschwindigkeitssensor(objekt.getPid());
    }

    private Windrichtungssensor erzeugeWindrichtungssensor(SystemObject objekt) {
        return new Windrichtungssensor(objekt.getPid());
    }

    private Umfelddatensensor erzeugeUmfelddatensensor(SystemObject objekt) {
        return new Umfelddatensensor(objekt.getPid());
    }

}
