/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.portals;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.sys.funclib.debug.*;
import de.bsvrz.sys.funclib.operatingMessage.*;

import java.text.*;
import java.time.*;
import java.util.*;

/**
 * Sendet Verkehrszustände an den Datenverteiler.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("ClassFanOutComplexity")
public class VerkehrszustandSender extends Kommunikation {

    private final Debug log = Debug.getLogger();
    private final MessageSender messageSender = MessageSender.getInstance();
    private final ClientSenderInterface sender = new SenderOhneSendesteuerung();
    private final Map<Systemobjekt, LocalDateTime> letzteZeitstempel = new LinkedHashMap<>();

    public VerkehrszustandSender(ClientDavInterface verbindung) {
        super(verbindung);
    }

    public void anmelden(Collection<Systemobjekt> objekte) {
        Collection<SystemObject> messquerschnitte = konvertiereObjekte(objekte, obj -> obj instanceof Messquerschnitt);
        doAnmelden(messquerschnitte);
        log.config(MessageFormat.format("Als Sender von Verkehrszustand für {0} Messquerschnitte angemeldet.", messquerschnitte.size()));

        Collection<SystemObject> strassenteilsegmente = konvertiereObjekte(objekte, obj -> obj instanceof Strassenteilsegment);
        doAnmelden(strassenteilsegmente);
        log.config(MessageFormat.format("Als Sender von Verkehrszustand für {0} Straßenteilsegmente angemeldet.", strassenteilsegmente.size()));
    }

    private void doAnmelden(Collection<SystemObject> objects) {
        DataDescription dd = getDatenbeschreibung();
        try {
            getVerbindung().subscribeSender(sender, objects, dd, SenderRole.source());
        } catch (OneSubscriptionPerSendData e) {
            log.error("Sendeanmeldung fehlgeschlagen: " + e, Arrays.asList(objects, dd));
            throw new IllegalStateException("Sendeanmeldung fehlgeschlagen", e);
        }
    }

    private DataDescription getDatenbeschreibung() {
        return datenbeschreibung("atg.störfallZustand", "asp.störfallVerfahrenFuzzy");
    }

    public void sende(Collection<Datensatz<?, Stoerfallzustand>> datensaetze) {
        ResultData[] results = datensaetze.stream()
                .map(this::konvertiereDatensatz)
                .toArray(ResultData[]::new);
        try {
            getVerbindung().sendData(results);
            log.fine(MessageFormat.format("{0} Verkehrszustände gesendet", datensaetze.size()));
        } catch (SendSubscriptionNotConfirmed e) {
            String message = "Verkehrszustände konnten nicht gesendet werden: " + e;
            log.warning(message, e);
            messageSender.sendMessage(MessageType.SYSTEM_DOMAIN, MessageGrade.WARNING, message);
        }
    }

    private ResultData konvertiereDatensatz(Datensatz<?, Stoerfallzustand> datensatz) {
        SystemObject obj = getModell().getObject(datensatz.getSystemobjekt().getPid());
        AttributeGroup atg = getDatenbeschreibung().getAttributeGroup();
        long time = konvertiereZeitstempel(datensatz.getZeitstempel());
        Data data = konvertiereDatum(datensatz, getVerbindung().createData(atg));
        return new ResultData(obj, getDatenbeschreibung(), time, data);
    }

    private Data konvertiereDatum(Datensatz<?, Stoerfallzustand> datensatz, Data data) {
        Stoerfallzustand zustand = datensatz.getDatum();
        data.getTimeValue("T").setSeconds(bestimmeIntervalldauerInSekunden(datensatz));
        data.getUnscaledValue("Situation").setText(konvertiereSituation(zustand));
        data.getTimeValue("Horizont").setSeconds(0);
        data.getItem("Güte").getScaledValue("Index").set(zustand.getGueteindex().getWert());
        data.getItem("Güte").getUnscaledValue("Verfahren").setText("Standard");
        return data;
    }

    private long bestimmeIntervalldauerInSekunden(Datensatz<?, Stoerfallzustand> datensatz) {
        Duration d = Duration.ZERO;
        Systemobjekt objekt = datensatz.getSystemobjekt();
        if (letzteZeitstempel.containsKey(objekt)) {
            d = Duration.between(letzteZeitstempel.get(objekt), datensatz.getZeitstempel());
        }
        letzteZeitstempel.put(objekt, datensatz.getZeitstempel());
        return d.getSeconds();
    }

    private static String konvertiereSituation(Stoerfallzustand zustand) {
        String situation;
        switch (zustand.getSituation()) {
            case KEINE_AUSSAGE:
                situation = "keine Aussage";
                break;
            case FREIER_VERKEHR:
                situation = "freier Verkehr";
                break;
            case DICHTER_VERKEHR:
                situation = "dichter Verkehr";
                break;
            case ZAEHFLIESSENDER_VERKEHR:
                situation = "zähfließender Verkehr";
                break;
            case STAU:
                situation = "Stau";
                break;
            default:
                throw new IllegalStateException("[unreachable code] Nicht behandelter Störfallzustand: " + zustand.getSituation());
        }
        return situation;
    }

}
